#
# ShotCreate.py - The batch backend program that does all the interesting
#   work in creating the shot tables.
#
# Main Operations:
#
#    If requested - interpolates headings from navigation data 
#
#    For every shot:
#       Linearly interpolates the heading between the two nav headings (if not an exact
#       time match).
#
#       Gets the lat/long position - if using shot positions, uses them directly.  If
#       interpolating from nav positions, interpolates linearly between the two
#       bounding navigation positions (neighbors).
#
#       If requested - corrects for shot gun / gps offsets, given suitable
#       behind/right offsets.
#
#       Gets the range from the shot to the obs, with the appropriate sign assigned
#       through a simple heuristic..
#
#       Corrects for obs time drift, as well as gun delay and ship-obs static offsets
#
#
# This program must be supplied the filename of a configuration file describing the
# options we accept.  This file is edited by the shotGui program, and is in a .ini
# style format.  The "Options" class handles the loading of these options, and all
# options are available as a simple variable within an Options object (eg:
# option.variable).
#
# Dalhousie University - Department of Oceanography.
#
# Author: Dave LeBlanc, dleblanc@cs.dal.ca, October 2002.

import sys
from math import *
import re
from bisect import bisect_left

# Import our useful modules
from shottab.options.options import Options
import shottab.fileio.fileio as fileio
from shottab.heading import *
from shottab.Itpl15 import ItplError
import shottab.Itpl15

# Use the psyco optimizer (if available) to optimize cpu intensive functions
# TODO: psyco seems to screw up the regex methods. Disabled for now.
#try:
#    import psyco
#    psyco.profile(0.1, memory=200)
#except ImportError:
#    print("Warning: psyco optimizer not available")

M_TO_KM = 1000

# The version and last date of modification
VERSION = "1.1.0 - August 14, 2004"

# The structure that defines how we deal with data loaded from shot time
# and navigation files - 1st field is the match name that will be matched
# with the regular expression when matching a line, and the second field
# is a function to convert the string representation to an internal
# data type.  The order here is important, this is the index that
# the item will be at when referencing it, such as:
#   navData[3] -- wll be the 'sec' field.
# Please see fileio.py for more information.
navValueTypes = (("day", int), ("hour", int), \
                 ("min", int), ("sec", float), \
                 ("partialSec", str), \
                 ("lat", fileio.strToLatLon), \
                 ("lon", fileio.strToLatLon), \
                 ("heading", float))

shotTimeValueTypes = (("shotNo", int), ("day", int), \
                      ("hour", int), ("min", int), \
                      ("sec", float), ("partialSec", str),
                      ("lat", fileio.strToLatLon), \
                      ("lon", fileio.strToLatLon))

# The time fields are not shown above - they will be appended below.

NAV_DAY_IDX = 0
NAV_HOUR_IDX = 1
NAV_MIN_IDX = 2
NAV_SEC_IDX = 3
NAV_PARTIAL_SEC_IDX = 4
NAV_LAT_IDX = 5
NAV_LON_IDX = 6
NAV_HEADING_IDX = 7
NAV_TIME_IDX = len(navValueTypes)

SHOT_DAY_IDX = 1
SHOT_HOUR_IDX = 2
SHOT_MIN_IDX = 3
SHOT_SEC_IDX = 4
SHOT_PARTIAL_SEC_IDX = 5
SHOT_LAT_IDX = 6
SHOT_LON_IDX = 7
SHOT_TIME_IDX = len(shotTimeValueTypes)

NANOSECS_PER_SEC = long(1e9)
NANOSECS_PER_MIN = 60 * NANOSECS_PER_SEC
NANOSECS_PER_HOUR = 60 * NANOSECS_PER_MIN
NANOSECS_PER_DAY = 24 * NANOSECS_PER_HOUR

class TimeReader(fileio.LazyLineReader):
    """
    Simply extends fileio.LazyLineReader to append a time stamp to each
    record converted.  This will be used to load shot time data, and
    the nav data loader will also use this to append times.

    Shot data will use this class directly,
    
    Nav data will extend it to append interpolated heading if requested - if not they
    also use this class directly.

    use like:

      a = TimeReader(args...)
      a[0]
      a[1]

    where a[0], a[1] return a list corresponding to the lines of the file
    we converted from the text data - except the final element will be a long value
    representing the time in nanoseconds.
    """
    def __init__(self, regExp, valueTypes, fileName, skipLines = 0):
        fileio.LazyLineReader.__init__(self, regExp, valueTypes, fileName, skipLines)

        # Since nanoseconds are optional, explicitly allow this
        self.partialSecIdx = -1
        
        # Find the indexes of the time elements - could also hard code this
        # (this is particularly inelegant)
        i = 0
        for (item, type) in valueTypes:
            if item == "day":
                self.dayIdx = i
            elif item == "hour":
                self.hourIdx = i
            elif item == "min":
                self.minIdx = i
            elif item == "sec":
                self.secIdx = i
            elif item == "partialSec":
                self.partialSecIdx = i

            i += 1

    def appendTime(self, record):
        """
        Append the time to the given record (list) - since lists are mutable, we
        modify it in-place (dont return a new one)
        """
        if record:

            # If partialSecs found, truncate secs, and add the partial seconds
            secs = record[self.secIdx]
            if self.partialSecIdx >= 0 and record[self.partialSecIdx]:
                secs = int(secs) + float("0." + record[self.partialSecIdx])
                
            record.append(getNanoTime(record[self.dayIdx], record[self.hourIdx], 
                record[self.minIdx], secs))

    def __getitem__(self, itemNo):
        # Append the time to this record
        record = fileio.LazyLineReader.__getitem__(self, itemNo)

        # If record != None - append time (assume we have all fields except
        # nano - is allowed to be None), else leave as None.
        self.appendTime(record)
            
        return record

class InterpolateHeadingReader(TimeReader):
    """
    Extends ShotTimeReader (simply appends a time element to each item) to support
    interpolation of heading times through a 'running sum' approach.

    This requires being able to search back and forwards a certain number of elements -
    thus using the cache of LazyLineReader so as to not read the same lines multiple
    times and ensure sequential access.

    Because of this approach the first and last few elements will be invalidated - beacuse
    we either:
        - cant get a prior element to figure out the _current_ heading, or
        - cant get the required number of elements before and after this element to
           allow for our running sum interpolation.

    So we will adjust the 0th index here, and simply raise an exception if we read too far
    to allow proper interpolation.
    """

    # Maximum number of elements in a heading cache before we flush it.
    # big so we dont thrash too much memory
    # TODO: refactor this so we don't throw away useful data
    MAX_CACHE_SIZE = 100000

    SUM_NO = 9
    
    # relative indexes starting from 0.
    MID_IDX = int(ceil(SUM_NO / 2.0) - 1)            # offset to the 'current' item
    SIDE_COUNT = int(floor((SUM_NO - 1) / 2.0))      # number on either side of the current

    def __init__(self, regExp, valueTypes, fileName, latLonIdx, hdgIdx, \
                 centralMeridian, skipLines = 0):

        # Call parent's constructor
        TimeReader.__init__(self, regExp, valueTypes, fileName, skipLines)

        self.latLonIdx = latLonIdx
        self.hdgIdx = hdgIdx
        self.centralMeridian = centralMeridian

        # Cache both the 'current' heading (heading between current and last) and
        # 'interpolated' heading (running sum of a few before and after)
        # in a set of hashes (maps a value to an index)
        #
        # We access these as
        #    heading = curHeadingCache[recordNo]
        self.curHeadingCache = {}
        self.avgHeadingCache = {}


    def getHeading(self, idx):
        """
        Get a heading for a given index - starting at 1 or greater
        (index 0 has no prior entry, so we cannot get it's heading)

        This is the 'current' heading, not the averaged heading.
        """

        # If it's cached, simply use that
        if self.curHeadingCache.has_key(idx):
            return self.curHeadingCache[idx]

        # Not cached - need to derive it from the previous and current lat/long
        # corrdinates. (this is done in heading.py).
        prevLine = fileio.LazyLineReader.__getitem__(self, idx - 1)
        curLine = fileio.LazyLineReader.__getitem__(self, idx)
        
        (prevLat, prevLon) =  (prevLine[self.latLonIdx[0]], prevLine[self.latLonIdx[1]])
        (curLat, curLon) =  (curLine[self.latLonIdx[0]], curLine[self.latLonIdx[1]])

        heading = getHeading(prevLat, prevLon, curLat, curLon, self.centralMeridian)

        # Cache this entry
        self.curHeadingCache[idx] = heading
        
        # Return the heading
        return heading

    def __getitem__(self, itemNo):
        """
        Get the interpolated (averaged) heading for the given index.

        Since the first few records are invalid for an interpolated
        heading, index 0 actully corresponds to approximately
        file record index 5 - since the 0th index is used
        for the next record's current heading, and 1st to 4th are
        used for the running sum.

        To achieve this we simply add MID_IDX + 1 (5) to the itemNo.

        We use 2 caches here to avoid duplicating too many
        computations - one for the 'current' heading - just associates
        itemNo with the heading.  And we have one for the averaged
        heading - associates itemNo with the whole record to be
        returned.
        """

        # Adjust the index to be relative to our file-indexes (from 0)
        # such that we start at index 5 if SUM_NO is 9.
        itemNo += self.MID_IDX + 1

        # return the cached value if it is stored
        if self.avgHeadingCache.has_key(itemNo):
            return self.avgHeadingCache[itemNo]

        # Read the record we are going to augment
        record = fileio.LazyLineReader.__getitem__(self, itemNo)

        sum = 0.0
        absSum = 0.0
        # interpolate the heading
        try:
            for i in range(itemNo - self.SIDE_COUNT, itemNo + self.SIDE_COUNT + 1):
                hdg = self.getHeading(i)
                sum += hdg
                absSum += abs(hdg)
        except fileio.LazyEOF:
            # We are trying to average beyond our
            raise InterpolateFileEnd("read to end of file")

        # Calculate the running sum - divide the absolute sum by the
        # number of divisions, and follow up by fixing the sign.
        # Python doesnt have a dsign, so we simply do that here        
        heading = absSum / float(sumNo)       
        if sum < 0.0:
            heading = -heading

        # Update the heading field
        record[self.hdgIdx] = heading

        # If the caches grow too big - delete them so we dont allocate too much
        # memory when dealing with large files
        if len(self.curHeadingCache) > self.MAX_CACHE_SIZE:
            self.curHeadingCache = {}
            self.avgHeadingCache = {}

        # append the time count and return the record (called from a class context)
        TimeReader.appendTime(self, record)

        # Cache the record (whole record) for future accesses
        self.avgHeadingCache[itemNo] = record

        return record


class NeighborError(Exception):
    pass

class InterpolateFileEnd(Exception):
    pass

def getNavNeighborIdx(navData, time, lastIdx):
    """
    Find the two neighbors (time wise) to the given time value.

    If it is equal to either, then either valid neighbor could
    be returned, but one of the indexes would correspond to an
    exact match.  When interpolation is performed, it will simply
    receive the value of the equal neighbor (thus no special
    case needed).

      Now that LazyReader is being used - we cannot rely on an exact
      maximum length being known - but we will use len() for an
      approximaton.

      Please read fileio:LazyLineReader:seekApproxLine for information
      on approximate file seeking, and how that relates to this function
      (if we iterate linearly - no problem, if we bounce around in a large
      file - then it uses the approximate seek function) -- essentially,
      we cannot guarantee that the position that we seek to is exactly where
      we want to be if the line lengths are not all of equal length -- but this
      is ok, because we only want the neighboring times, and we don't care if
      the index we seek to is a little off. -- This allows us to seek around
      in extremely large files very quickly.

    If the 'current' position's time is within a given margin (200 seconds)
    of the intended shot time, we simply do a linear search (no potential
    line skipping inaccuracies with this process, otherwise we do the following:

    The algorithm used for this function is a 'binary search' - meaning
    we start with the middle element, and go to the left or right sub
    list on the value of this element compared to our target time.
    This requires our data to be sorted according to time (which nav data
    should always be).

    We start at the middle element, and go to whatever sub-section is closer
    to our intended time.

    This binary-search process is (is an algorithm analysis sense) far more
    efficient than the linear search.  It is O(lg n) which means as n (our
    input size) grows, the execution time grows proportionally to lg n, while
    a linear search is O(n) - which is far worse, especially for large files
    where we may have to search through a very large amount of leading data to
    get to what we are interested in.
    
    """
    
    # The maximum time difference between target time and last indexed time
    # to allow for a simple linear search.  (200 seconds max)
    LINEAR_SEARCH_MAX_NANOS = long(200e9)


    # examine the 'current' element, if close enough - just perform a linear
    # search (this will comprise most cases).  If we need a backwards search,
    # perform this with binary search.
    i = lastIdx    
    if time - navData[i][NAV_TIME_IDX] <= LINEAR_SEARCH_MAX_NANOS:
        while navData[i + 1][NAV_TIME_IDX] < time:
            i += 1
        return (i, i+1)

    # else perform a binary search, where the returned indexes will
    # have (before, after) where before is <= time, and after is
    # after the time. -- adapted from bisect.py in the python distro.
    hi = len(navData)
    lo = 0
    while lo < hi:
        mid = (lo + hi) / 2
        if navData[mid][NAV_TIME_IDX] < time:
            lo = mid + 1
        else:
            hi = mid

    # Since the above gives the position we wish to _insert_ (if we
    # were to do so), we need to adjust so it gives neighbor indexes
    # (if at 0, we are ok, else adjust down by one)
    if lo > 0:
        lo -= 1

    # check if we haven't got a value greater (or less) than
    # the target value (outside navigation times) and raise exception.
    if navData[lo][NAV_TIME_IDX] > time or \
       navData[lo + 1][NAV_TIME_IDX] < time:
        raise NeighborError("no neighbors to shot time.")

    return (lo, lo + 1)
    

def getNanoTime(days, hours, mins, secs):
    """
    Return a unified time format - in nanoseconds
    Only to be used internally for comparisons and such.
    Uses a long type (postfix L), and can have unbounded range.

    Also supports conversion in the other direction (back to
    components).

    The advantage to this approach is absolutely no loss of precision.
    """
    return long(days * NANOSECS_PER_DAY + hours * NANOSECS_PER_HOUR + \
                + mins * NANOSECS_PER_MIN + secs * NANOSECS_PER_SEC)

def nanoToDayHourMinSec(nanoSecs):
    """
    Converts from a nanosecond time format to a Day, Hours, Minute, Second,
    format.
    """
    # NOTE: this is fairly inefficient, but somewhat simple.

    days = int(nanoSecs / NANOSECS_PER_DAY)
    nanoSecs -= days * NANOSECS_PER_DAY

    hours = int(nanoSecs / NANOSECS_PER_HOUR)
    nanoSecs -= hours * NANOSECS_PER_HOUR

    mins = int(nanoSecs / NANOSECS_PER_MIN)
    nanoSecs -= mins * NANOSECS_PER_MIN

    # NOTE: secs is a float value.
    secs = float(nanoSecs) / NANOSECS_PER_SEC

    return (days, hours, mins, secs)

def nanoToStr(nanoSecs):
    dhms = nanoToDayHourMinSec(nanoSecs)
    return "day %d, hour %d, min %d, sec %.4f" % \
           (dhms[0], dhms[1], dhms[2], dhms[3])

def correctGunGpsOffset(lat, lon, heading, behind, right, \
        centralMeridian):
    """
    A function which returns (lat, lon) of the corrected position with
    respect to the shot gun - gps offset and current heading.

    Converts to UTM coordinates, adds the appropriate X and Y components
    of the gun offset to the northing/easting respectively, and converts
    these coordinates back to Lat/Long, and returns these.
    """
    headingRad = heading / (180 / pi)

    (north, east) = latLongToUTM(lat, lon, centralMeridian)

    north = north - behind * cos(headingRad) + right * sin(headingRad)
    east = east - behind * sin(headingRad) - right * cos(headingRad)

    # (returns a lat/lon tuple)
    return UTMToLatLong(north, east, centralMeridian)



(_prevRangeNorth, _prevRangeEast) = (0.0, 0.0)
_prevRange = 0.0
(_prevGoNorth, prevGoEast) = (0, 0)
_prevToward = 0
def getRange(lat, lon, obsLat, obsLon, centralMeridian, shotIter):
    # NOTE: I'm storing the previous north/east range values in
    # (global) _prevRangeNorth and _prevRangeEast variables.
    #

    global _prevRangeNorth, _prevRangeEast, _prevRange, \
           _prevToward, _prevGoEast, _prevGoNorth
    
    (goNorth, goEast, toward) = (0, 0, 0)

    (north, east) = latLongToUTM(lat, lon, centralMeridian)
    (obsNorth, obsEast) = latLongToUTM(obsLat, obsLon, centralMeridian)
    
    rangeEast = east - obsEast
    rangeNorth = north - obsNorth

    # compute the range
    obsRange = sqrt(rangeEast**2 + rangeNorth**2) / M_TO_KM

    # Now compute the sign of the obs range
    toward = abs(obsRange) <= abs(_prevRange)     
    goEast = rangeEast > _prevRangeEast
    goNorth = rangeNorth > _prevRangeNorth

    if shotIter == 0:
        # 1st shot only
        if rangeEast <= -1:
            obsRange = -obsRange
        elif abs(rangeEast) < 1 and rangeNorth < 0:
            obsRange = -obsRange

    elif shotIter == 1:
        # 2nd shot
        if _prevRange < 0:
            obsRange = -obsRange

    else:
        # For the later shots

        if _prevToward and not toward and \
               ((abs(rangeEast) > abs(rangeNorth) and _prevGoEast == goEast) or \
               (abs(rangeNorth) >= abs(rangeEast) and _prevGoNorth == goNorth)):
            if _prevRange > 0:
                obsRange = -obsRange

            # Else keep it positive
            
        else:
            if _prevRange < 0:
                obsRange = -obsRange                


    (_prevRangeNorth, _prevRangeEast) = (rangeNorth, rangeEast)
    _prevRange = obsRange
    (_prevGoNorth, _prevGoEast) = (goNorth, goEast)
    _prevToward = toward
    return obsRange

def printProgress(progressAmount):
    """Simply print the progress to the screen"""
    print("%2d%s complete" % (int(progressAmount * 100), r"%"))
    
# The main routine
def shottabMain(optionsFile, progressMethod=printProgress):
    """
    This is the main routine for converting the Shotlog and Navigation input into
    an output Shotlog Table suitable for input into the DOBS segy conversion program.

    Options are loaded from the file supplied on the command line (.ini format). -
    These are loaded into the 'options' object, and the individual options are
    available as variables of this object - eg 'options.interpolatePositions'.

    The NavData and ShotTimeData objects are indexable - eg: navData[3], and give
    the converted value at that specific line number (minus header lines and
    if the navigation data was interpolated, this removes the first few).  This returns
    a list of items that were converted such as (for an example nav data line):

      [1, 235, 21, 3, 40.0, 42.123, -62.123, 12345678]

    Which correspond respectively to:

      shotNo, day, hour, minute, second, latitude, longitude, nanoSeconds

    With each item having the type (int, float, etc) as specified in the given
    'valueTypes' specified when creating the NavData/ShotTimeData object.
    Note that the last field is computed on the fly from the other time fields.
    
    """

    try:
        options = Options(optionsFile)
    except IOError, err:
        print("Unable to open configuration file: %s" % str(err))
        sys.exit(1)

    try:
        shotOutputFile = open(options.outputShotlogName, "w+")
    except IOError, err:
        print("Unable to open shottab output file: %s" % str(err))
        sys.exit(1)


    # Attempt to initialize our data loadings objects - catch errors if it
    # fails.
    try:
        
        # stdout used so it doesnt put out a newline.  Flush needed to immediately
        # show output.
        sys.stdout.write("Loading navigation data...")
        sys.stdout.flush()
        
        whichData = "navigation data"
        if options.interpolateHeadings:
            # Make navData an object which reads it's data and
            # interpolates headings (and appends a time stamp)
            # -- Removes the first few lines for interpolation.
            navData = InterpolateHeadingReader(options.navRegExp, navValueTypes, \
                                               options.navFile, \
                                               (NAV_LAT_IDX, NAV_LON_IDX), \
                                               NAV_HEADING_IDX, \
                                               options.centralMeridian, \
                                               skipLines = options.navSkipLines)
        else:
            # Simply make navData an object that appends a time stamp to it's read
            # data
            navData = TimeReader(options.navRegExp, navValueTypes, \
                                 options.navFile, \
                                 skipLines = options.navSkipLines)

        print("done")

        whichData = "shot time data"
        sys.stdout.write("Loading Shot Time data...")
        sys.stdout.flush()    
        shotTimeData = TimeReader(options.shotTimeRegExp, shotTimeValueTypes, \
                                  options.shotTimeFile, \
                                  skipLines = options.shotTimeSkipLines)

        print("done")

    except re.error, errMsg:
        print("Regular expression error in %s: %s" %(whichData, str(errMsg)))
        sys.exit(1)
        
    except IOError, errMsg:
        print("I/O error in loading %s: %s" %(whichData, str(errMsg)))
        sys.exit(1)
        
    except ItplError, errMsg:
        print("Function-expansion error in %s: %s" %(whichData, str(errMsg)))
        sys.exit(1)

    sys.stdout.write("Processing data...\n")
    sys.stdout.flush()

    # Find out the drift rate
    day, hour, min, sec = fileio.strToDate(options.obsClockResetTime)
    obsResetTime = getNanoTime(day, hour, min, sec)
    day, hour, min, sec = fileio.strToDate(options.obsCalibrationTime)
    obsCalibTime = getNanoTime(day, hour, min, sec)
    driftRate = options.clockDrift / (obsCalibTime - obsResetTime)

    # Don't pollute our variable space after here
    del day, hour, min, sec

    # Print the header
    shotOutputFile.write(options.cruiseID + "\n")
    shotOutputFile.write("%12.6f%12.6f  OBS Ref. latitude and longitude\n" \
                         %(options.obsLat, options.obsLon))
    
    clockReset = fileio.strToDate(options.obsClockResetTime)
    shotOutputFile.write("%3d%3d%3d%3.0f.         Time of clock reset\n" \
                         %(clockReset[0], clockReset[1], clockReset[2], clockReset[3]))

    shotOutputFile.write("     .0000            Clock drift rate (ms/hr)\n" +
                         "    0                 Delay in firing gun (interger msec)\n" +
                         "    0                 Offset(interger msec):obs&ship clocks\n")
    shotOutputFile.write(" SHOT DAY HR MM     SEC   LATITUDE   LONGITUDE    RANGE\n")

    # Main loop - process each line of shot data
    beforeIdx = 0
    i = 0
    while 1:
        try:
            shot = shotTimeData[i]

            if not shot:
                print("Warning: could not parse line %d, skipping." % (i + 1))
                continue

            # Get the position and heading for this shot - if using nav
            # positions, interpolate linearly between the two boundary
            # navigation times relative to the shotlog time (both
            # position and heading).
            # 
            # If using shot positions, no logic required for positon,
            # just use the position from the shotlog data, but use
            # linear interpolation for the heading from the nav data.
            
            time = shot[SHOT_TIME_IDX]
            
            # Get the neighboring nav times for this shot
            # TODO: if this raises NeighborError, tell the file line no and
            #   search shot time (in day hr min fmt).
            try:
                (beforeIdx, afterIdx) = getNavNeighborIdx(navData, time, beforeIdx)
            except NeighborError, errMsg:
                print("Nav Neighbor Finding Error: %s" % str(errMsg))
                print("\tat time day: %s" % nanoToStr(time))
                sys.exit(1)

        except InterpolateFileEnd:
            print "file ended at index %d" % i
            break
        
        except fileio.LazyEOF:
            # we are at the end of the file
            print "file end at index %d" % i
            break

        # Get the neighboring headings and times
        beforeRec = navData[beforeIdx]
        afterRec = navData[afterIdx]
        (beforeHdg, afterHdg) = (beforeRec[NAV_HEADING_IDX], \
                                 afterRec[NAV_HEADING_IDX])
        (beforeTime, afterTime) = (beforeRec[NAV_TIME_IDX], \
                                   afterRec[NAV_TIME_IDX])

        if (beforeTime == time):
            heading = beforeHdg
        elif (afterTime == time):
            heading = afterHdg
        else:
            # Interpolate heading
            heading = beforeHdg + (afterHdg - beforeHdg) / (afterTime - beforeTime) * \
                (time - beforeTime)

        # Get the lat/lon position
        if options.interpolatePositions:
            # Interpolate lat/lon position
            beforeLat, afterLat = (beforeRec[NAV_LAT_IDX], \
                                   afterRec[NAV_LAT_IDX])
            beforeLon, afterLon = (beforeRec[NAV_LON_IDX], \
                                   afterRec[NAV_LON_IDX])
            lat = beforeLat + (afterLat - beforeLat) / (afterTime - beforeTime) * \
                  (time - beforeTime)
            lon = beforeLon + (afterLon - beforeLon) / (afterTime - beforeTime) * \
                  (time - beforeTime)            
        else:
            # Use the positions directly from the shot data
            lat = shot[SHOT_LAT_IDX]
            lon = shot[SHOT_LON_IDX]

        # (optionally) correct for gun-gps offset
        if options.correctForOffset:
            (lat, lon) = correctGunGpsOffset(lat, lon, heading, \
                 options.gunOffsetBehind, \
                 options.gunOffsetRight, \
                 options.centralMeridian)

        # Get the range to the obs
        obsRange = getRange(lat, lon, options.obsLat, options.obsLon, \
            options.centralMeridian, i)

        # If we are to make it positive regardless - do this.
        if options.makeRangePositive:
            obsRange = abs(obsRange)
            
        # correct the shot time for time delay/drift (in milliseconds)
        # driftRate is milliseconds drift per time in nanoseconds
        driftMs = (time - obsResetTime) * driftRate

        # Add the time offsets - these are in milliseconds (msec - 1e3)
        # so we multiply by 1000000 (1e6) to get nanoseconds - 1e9.
        time += long((options.gunDelay + driftMs + options.obsShipOffset) * 1e6)

        (day, hour, min, sec) = nanoToDayHourMinSec(time)

        # Write out this line to the file        
        shotOutputFile.write("%5d %3d %2d %2d %7.4f %10.6f %11.6f %8.3f\n" % \
              (shot[0], day, hour, min, sec, lat, lon, obsRange))

        if i % 200 == 0:
            # Update the progress every so often
            progressMethod(shotTimeData.getRelativeCompletion())

        i += 1

    shotOutputFile.close()
    print("done")
