#!/usr/bin/env python
#
# This is the GUI editor of the options file for the pyShottab program.
#
# This is built with wxPython - which is a binding of the cross-platform
#  wxWindows GUI toolkit for python (provides seamless Windows/Unix GUI
#  transparency).  Please see www.wxpython.org for info and documentation.
#
# This dialog simply edits the values of the Options object (self.options)
# which we save/load from a given file (in .ini format).  This is used to
# pass options to the batch shottab program (which also loads using the
# Options object, such that duplicated I/O routines are kept to a minimum).
# Please see options.py for more information.
#
# We also provide a 'run' button, which simply runs the shotCreate
# command-line program (the workhorse of this package) and displays
# the output in a TextViewDialog.
#
# GUI CONCEPT - put a bunch of controls on, let the user edit them - do
# not intercept most events, just grab the value when the save button
# is pressed, and use these values to save to a file.  This way we only
# need a few event handlers.  updateGui() is called to update the state
# of the gui to the state of the self.options object.
#
#
# A variety of controls (some custom) are used to ensure that the input is
# valid (see guiComponents for these).  These are:
#
#    IntEditCtrl - a custom control for editing Int's - doesnt allow
#      any other input
#
#    FloatEditCtrl - same as IntEditCtrl but for Floats.
#
#    DateCtrl - Control for editing date values with day, hour, minute as int,
#      and second as a float.
#
#    TextDataDialog - A large dialog which allows the description of a regular
#      expression (text matching scheme) to edit the input of text-data into
#      a form which our program can use.  This same dialog is used for the
#      input of both Shot Time and Navigation data, with slightly different
#      parameters passed in to customize it.
#      This dialog also provides a small space to test the input, allowing a
#      user to refine their regular-expression quickly.
#
#    TextViewDialog - Simply displays some text with an OK/Cancel button.
#
#
# 
#
# Dalhousie University - Department of Oceanography.
#
# Author: Dave LeBlanc, dleblanc@cs.dal.ca, October 2002.

from wxPython.wx import *
import sys
import re
import os.path
import ConfigParser
import time
import threading

# Import our useful modules
from shottab.options import options as options
from shottab.fileio import fileio as fileio
from guiComponents import *
from shottab.core.shotCreate import shotTimeValueTypes, navValueTypes, VERSION

import shottab.help as help
import shottab.core.shotCreate as shotCreate

# A couple events that we use to communicate from our run thread back to the UI
RUN_COMPLETE_EVENT_ID = wxNewId()
RUN_ADD__OUTPUT_EVENT_ID = wxNewId()

class MainWindow(wxFrame):
    """
    This is the main window of the Options editor.
    """
    
    # Menu ID's   (these are not normal variables - these are shared
    # between all objects of the same class: called 'class variables')
    SHOT_OUT_FILE_BUTTON_ID = wxNewId()
    SHOT_TIME_BUTTON_ID = wxNewId()
    NAV_BUTTON_ID = wxNewId()
    RUN_BUTTON_ID = wxNewId()
    
    OPEN_MENU_ID = wxNewId()
    SAVE_MENU_ID = wxNewId()
    SAVE_AS_MENU_ID = wxNewId()
    EXIT_MENU_ID = wxNewId()
    HELP_MENU_ID = wxNewId()
    ABOUT_MENU_ID = wxNewId()
    
    def __init__(self, parent, id, title, cfgFileName=None):
        # Call parent's constructor with appropriate values
        wxFrame.__init__(self, parent, id, title, \
                         style=wxDEFAULT_FRAME_STYLE)

        # The name of our configuration file.
        # If it is None, then we havent provided one yet (is ok - just creates
        # attributes with defaults)
        self.cfgFileName = cfgFileName

        # Create our options object.  If passed file name is 'None', we simply create a
        # prototypical options object with suitable defaults.

        try:
            self.options = options.Options(self.cfgFileName)

        # If any errors, just create options object with defaults and continue
        except IOError, errMsg:
            dlg = wxMessageDialog(self, "I/O Error loading options file: %s" % str(errMsg), \
                                  "I/O Error Loading Options File", wxOK | wxICON_WARNING) 
            dlg.ShowModal()
            dlg.Destroy()

            self.options = options.Options(None)
            
        except ConfigParser.Error, errMsg:
            dlg = wxMessageDialog(self, "Error loading options data: %s" % str(errMsg), \
                                  "Options Loading Error", wxOK | wxICON_WARNING) 
            dlg.ShowModal()
            dlg.Destroy()

            self.options = options.Options(None)

        # Make a grid layout - labels on the left, edit controls on the right.
        # Items are placed in left to right, top down order (starting a new row after
        # every 2 items).
        self.sizer = wxFlexGridSizer(cols = 2, vgap = 2)

        mainPanel = wxPanel(self, -1, style=wxSIMPLE_BORDER)

        # Designate row 1 (indexed from 0 - therefore the right column) as growable.
        self.sizer.AddGrowableCol(1)


        # This is the Cruise ID edit box - with a small top border
        self.cruiseText = wxTextCtrl(mainPanel, -1, value=self.options.cruiseID)
        self.sizer.Add(wxStaticText(mainPanel, -1, "Cruise ID:"), 0, wxTOP, 5)
        self.sizer.Add(self.cruiseText, 1, wxEXPAND | wxTOP, 5)



        # Add a label
        self.sizer.Add(wxStaticText(mainPanel, -1, "Output Shot Table File Name:"), 0)

        shotlogFileSizer = wxBoxSizer(wxHORIZONTAL)
        # A label for the filename
        self.shotlogText = wxTextCtrl(mainPanel, -1, style = wxTE_READONLY, \
                                      value=self.options.outputShotlogName)
        self.shotlogText.SetBackgroundColour(wxLIGHT_GREY)
        
        # A button to choose
        shotlogFileButton = wxButton(mainPanel, self.SHOT_OUT_FILE_BUTTON_ID, "Choose...")
        EVT_BUTTON(self, self.SHOT_OUT_FILE_BUTTON_ID, self.chooseShotFile)

        # Add both elements to the horizontal sizer
        shotlogFileSizer.Add(self.shotlogText, 1, wxEXPAND)        
        shotlogFileSizer.Add(shotlogFileButton, 0)

        # And the horizontal sizer
        self.sizer.Add(shotlogFileSizer, 1, wxEXPAND)



        # The shotlog options (use from file, interpolate)
        self.sizer.Add(wxStaticText(mainPanel, -1, "Shot Positions:"), 0)        
        self.positionsRadio = wxRadioBox(mainPanel, -1, "", \
                                    choices = ["use shot positions from shotlog data", \
                                               "interpolate shot positions from nav data"], \
                                    majorDimension = 1, \
                                    style = wxRA_SPECIFY_COLS)

        self.sizer.Add(self.positionsRadio, 1, wxEXPAND)        

        # Shot Time dialog button...
        shotTimeButton = wxButton(mainPanel, self.SHOT_TIME_BUTTON_ID, \
                                  "Describe Shotlog Data")
        EVT_BUTTON(self, self.SHOT_TIME_BUTTON_ID, self.shotTimeClicked)        
        self.sizer.Add(wxStaticText(mainPanel, -1, "Shotlog Data:"), 0)        
        self.sizer.Add(shotTimeButton, 0)        

        # Nav Data dialog button (same as above but for nav data)...
        navButton = wxButton(mainPanel, self.NAV_BUTTON_ID, "Describe Navigation Data")
        EVT_BUTTON(self, self.NAV_BUTTON_ID, self.navClicked)        
        self.sizer.Add(wxStaticText(mainPanel, -1, "Navigation Data:"), 0)        
        self.sizer.Add(navButton, 0)        


        # Interpolate headings from nav data
        self.sizer.Add(wxStaticText(mainPanel, -1, "Heading Interpolation:"), 0)        
        self.interpolateHeadingsCheck = wxCheckBox(mainPanel, -1, \
                                              "Interpolate headings from nav data")
        self.sizer.Add(self.interpolateHeadingsCheck)
        
        # The 'correct shot positions for offset' option
        self.sizer.Add(wxStaticText(mainPanel, -1, "Shot Position Offset:"), 0)        
        self.correctShotPosCheck = wxCheckBox(mainPanel, -1, \
                                              "correct shot positions for gun-gps offset" + \
                                              " (requires headings)")
        self.sizer.Add(self.correctShotPosCheck)


        # Gun-antenna offset
        self.sizer.Add(wxStaticText(mainPanel, -1, "Shot Position Offset:"), 0)

        gunOffSizer = wxBoxSizer(wxHORIZONTAL)
        self.gunOffsetBehind = FloatEditCtrl(mainPanel, \
                                           value=self.options.gunOffsetBehind)
        self.gunOffsetRight = FloatEditCtrl(mainPanel,
                                          value=self.options.gunOffsetRight)

        gunOffSizer.Add(wxStaticText(mainPanel, -1, "Behind (meters):"), 0)
        gunOffSizer.Add(self.gunOffsetBehind)
        gunOffSizer.Add(wxStaticText(mainPanel, -1, "Right (meters):"), 0)        
        gunOffSizer.Add(self.gunOffsetRight)
        
        self.sizer.Add(gunOffSizer, 1, wxEXPAND)


        # Make range positive always if checked 
        self.sizer.Add(wxStaticText(mainPanel, -1, "Range Sign:"), 0)        
        self.makeRangePositive = wxCheckBox(mainPanel, -1, \
                                           "Always make output range sign positive")
        self.sizer.Add(self.makeRangePositive)


        # OBS Info follows

        
        # OBS position
        self.sizer.Add(wxStaticText(mainPanel, -1, "OBS Position:"), 0)
        self.obsCoord = LatLonCtrl(mainPanel, \
                                   value=(self.options.obsLat, self.options.obsLon))
        self.sizer.Add(self.obsCoord, 0)


        
        self.sizer.Add(wxStaticText(mainPanel, -1, "Central Meridian:"), 0)
        self.centralMeridian = IntEditCtrl(mainPanel, \
                                             value=self.options.centralMeridian)
        self.sizer.Add(self.centralMeridian, 0)


        # OBS clock reset
        self.sizer.Add(wxStaticText(mainPanel, -1, "Time of clock reset:"), 0)
        self.dateReset = DateCtrl(mainPanel, \
                                  value=fileio.strToDate(self.options.obsClockResetTime))
        self.sizer.Add(self.dateReset, 1, wxEXPAND)
      
        # Offset betwen ship and OBS clock
        self.sizer.Add(wxStaticText(mainPanel, -1, "Offset between ship and OBS (msec):"), 0)
        self.obsShipOffsetEdit = FloatEditCtrl(mainPanel, \
                                               value=self.options.obsShipOffset)
        self.sizer.Add(self.obsShipOffsetEdit, 0)


        # OBS time of drift calibration
        self.sizer.Add(wxStaticText(mainPanel, -1, "Time of drift calibration:"), 0)
        self.dateCalib = DateCtrl(mainPanel, \
                                  value=fileio.strToDate(self.options.obsCalibrationTime))
        self.sizer.Add(self.dateCalib, 1, wxEXPAND)


        # Total clock drift in msec
        self.sizer.Add(wxStaticText(mainPanel, -1, "Total clock drift (msec):"), 0)
        self.clockDriftEdit = FloatEditCtrl(mainPanel, value=self.options.clockDrift)
        self.sizer.Add(self.clockDriftEdit, 0)
        

        # Gun delay in msec's
        self.sizer.Add(wxStaticText(mainPanel, -1, "Gun Delay (msec):"), 0)
        self.gunDelayEdit = FloatEditCtrl(mainPanel, value=self.options.gunDelay)
        self.sizer.Add(self.gunDelayEdit, 0)


        # Put a few buttons on the lower-right (save shares ID from menu)
        mainButtonGroup = wxBoxSizer(wxVERTICAL)
        
        saveButton = wxButton(mainPanel, self.SAVE_MENU_ID, "Save...")
        runButton = wxButton(mainPanel, self.RUN_BUTTON_ID, "Run...")
        
        EVT_BUTTON(self, self.SAVE_MENU_ID, self.menuSave)
        EVT_BUTTON(self, self.RUN_BUTTON_ID, self.runClicked)

        mainButtonGroup.Add(saveButton)
        mainButtonGroup.Add(runButton)        
        
        self.sizer.Add(wxWindow(mainPanel, -1), 0)  # Empty cell        
        self.sizer.Add(mainButtonGroup, 0, wxALIGN_RIGHT)


        # Add a set of menus
        menu = wxMenu()
        menu.Append(self.OPEN_MENU_ID, "&Open")
        menu.Append(self.SAVE_MENU_ID, "&Save")
        menu.Append(self.SAVE_AS_MENU_ID, "Save &As")
        menu.Append(self.EXIT_MENU_ID, "E&xit")
        # Set the event handlers for the above menus
        EVT_MENU(self, self.OPEN_MENU_ID, self.menuOpen)
        EVT_MENU(self, self.SAVE_MENU_ID, self.menuSave)
        EVT_MENU(self, self.SAVE_AS_MENU_ID, self.menuSaveAs)
        EVT_MENU(self, self.EXIT_MENU_ID, self.menuExit)


        # Add a help menu
        helpMenu = wxMenu()
        helpMenu.Append(self.HELP_MENU_ID, "&Help")
        helpMenu.Append(self.ABOUT_MENU_ID, "&About")
        EVT_MENU(self, self.HELP_MENU_ID, self.menuHelp)
        EVT_MENU(self, self.ABOUT_MENU_ID, self.menuAbout)
        
        menuBar = wxMenuBar()
        menuBar.Append(menu, "&File")
        menuBar.Append(helpMenu, "&Help")
        
        # Designate this as the frame's menu bar
        self.SetMenuBar(menuBar)

        # Register the (grid) sizer, and show our gui
        mainPanel.SetAutoLayout(true)
        mainPanel.SetSizer(self.sizer)
        
        # self.sizer.Fit(mainPanel)

        # Size the dialog to an (automatic) suitable default
        newSize = self.sizer.Fit(self)
        newSize[1] += 10
        self.SetSize(newSize)
        self.Centre()
        
        # Have the shottab run thread tell us when he's done
        self.Connect(-1, -1, RUN_COMPLETE_EVENT_ID, self.runComplete)
        self.Connect(-1, -1, RUN_ADD__OUTPUT_EVENT_ID, self.addRunOutput)

        # Sync out GUI state with the options state
        self.updateGui()
        
        self.Show(true)

    def updateGui(self):
        """Update the state of the gui from the self.options object"""
        self.cruiseText.SetValue(self.options.cruiseID)
        self.shotlogText.SetValue(self.options.outputShotlogName)
        self.positionsRadio.SetSelection(bool(self.options.interpolatePositions))
        self.interpolateHeadingsCheck.SetValue(bool(self.options.interpolateHeadings))
        self.correctShotPosCheck.SetValue(bool(self.options.correctForOffset))
        self.makeRangePositive.SetValue(bool(self.options.makeRangePositive))
        self.gunOffsetBehind.SetValue(self.options.gunOffsetBehind)
        self.gunOffsetRight.SetValue(self.options.gunOffsetRight)

        self.obsCoord.SetValue((self.options.obsLat, self.options.obsLon))
        self.centralMeridian.SetValue(self.options.centralMeridian)
        self.dateReset.SetValue(fileio.strToDate(self.options.obsClockResetTime))
        self.dateCalib.SetValue(fileio.strToDate(self.options.obsCalibrationTime))
        self.clockDriftEdit.SetValue(self.options.clockDrift)
        self.gunDelayEdit.SetValue(self.options.gunDelay)
        self.obsShipOffsetEdit.SetValue(self.options.obsShipOffset)
        
        # update shotTime/nav infos as well (this is not gui, but needs to be sync'd)
        self.shotTimeInfo = (self.options.shotTimeFile,
                             self.options.shotTimeSkipLines,
                             self.options.shotTimeRegExp)
        
        self.navInfo = (self.options.navFile,
                             self.options.navSkipLines,
                             self.options.navRegExp)
        

    def menuOpen(self, evt):
        fileDialog = wxFileDialog(self, "Choose Options Input File", style=wxOPEN)
        result = fileDialog.ShowModal()

        if result == wxID_OK:

            try:
                self.options = options.Options(fileDialog.GetPath())

                # If it worked, update our file name
                self.cfgFileName = fileDialog.GetPath()
                
            # If any errors, just create options object with defaults and continue
            except IOError, errMsg:
                dlg = wxMessageDialog(self, "I/O Error loading options file: %s" % str(errMsg), \
                                      "I/O Error Loading Options File", wxOK | wxICON_WARNING) 
                dlg.ShowModal()
                dlg.Destroy()

                self.options = options.Options(None)
            
            except ConfigParser.Error, errMsg:
                dlg = wxMessageDialog(self, "Error loading options data: %s" % str(errMsg), \
                                      "Options Loading Error", wxOK | wxICON_WARNING) 
                dlg.ShowModal()
                dlg.Destroy()

                self.options = options.Options(None)

            # Refresh the GUI state
            self.updateGui()



    def menuSaveAs(self, evt, pickOnly=0):
        """Prompt for a file name, update the self.cfgFileName variable, and call menuSave()"""
        fileDialog = wxFileDialog(self, "Choose Config Output File", style=wxSAVE)
        result = fileDialog.ShowModal()

        if result == wxID_OK:
            self.cfgFileName = fileDialog.GetPath()

            # forward to our save event handler if we are to save as well
            if not pickOnly:
                self.menuSave(evt)
        
        fileDialog.Destroy()
        return result


    def menuSave(self, evt):
        """
        Simply grab the values from the GUI and update the Options object,
        and use this to update the file.
        """

        # Ensure all the options are valid - don't allow a save until they are

        # If we did not pick a file - pick one now.
        if self.cfgFileName is None:
            if self.menuSaveAs(evt, pickOnly=1) != wxID_OK:
                # If they cancelled, simply exit this routine
                return

        # ensure both text file options are filled out
        if self.shotTimeInfo[0] == "" or self.shotTimeInfo[2] == "":
            dlg = wxMessageDialog(self, "Error: Shot Time information incomplete", \
                                  "Shot Time Incomplete", wxOK | wxICON_WARNING) 
            dlg.ShowModal()
            dlg.Destroy()
            return

        if self.navInfo[0] == "" or self.navInfo[2] == "":            
            dlg = wxMessageDialog(self, "Error: Navigation information incomplete", \
                                  "Navigation Incomplete", wxOK | wxICON_WARNING) 
            dlg.ShowModal()
            dlg.Destroy()
            return


        # Options all appear to be (as far a we can tell) valid
        # Commit any set options to the options object from the gui
        self.options.cruiseID = self.cruiseText.GetValue()
        self.options.outputShotlogName = self.shotlogText.GetValue()


        self.options.interpolatePositions = bool(self.positionsRadio.GetSelection())
        self.options.interpolateHeadings = bool(self.interpolateHeadingsCheck.GetValue())

        (self.options.shotTimeFile, self.options.shotTimeSkipLines, \
         self.options.shotTimeRegExp) =  self.shotTimeInfo

        (self.options.navFile, self.options.navSkipLines, \
         self.options.navRegExp) = self.navInfo

        self.options.correctForOffset = bool(self.correctShotPosCheck.GetValue())

        self.options.gunOffsetBehind = self.gunOffsetBehind.GetValue()
        self.options.gunOffsetRight = self.gunOffsetRight.GetValue()

        self.options.makeRangePositive = bool(self.makeRangePositive.GetValue())

        (self.options.obsLat, self.options.obsLon) = self.obsCoord.GetValue()

        self.options.centralMeridian = self.centralMeridian.GetValue()

        # Note that this is a tuple of day, hour, minute, second
        self.options.obsClockResetTime = str(self.dateReset.GetValue())
        self.options.obsCalibrationTime = str(self.dateCalib.GetValue())

        self.options.clockDrift = self.clockDriftEdit.GetValue()
        self.options.gunDelay = self.gunDelayEdit.GetValue()
        self.options.obsShipOffset = self.obsShipOffsetEdit.GetValue()

        # Save the file in it's cfg file format
        self.options.save(self.cfgFileName)

    def menuExit(self, evt):
        """Exit the program."""
        self.Close(true)
    
    def menuHelp(self, evt):
        """Show some help when the user clicks Help->Help.  Just using our TextViewDialog
        for now (Instead of the built in help classes). """
        dialog = TextViewDialog(self, "PyShottab Help", help.shotGuiHelp)
        dialog.ShowModal()
        dialog.Destroy()

    def menuAbout(self, evt):
        aboutDialog = wxMessageDialog(self, help.aboutShotGui,
                                      caption="About PyShottab", style=wxOK | wxCENTRE)
        aboutDialog.ShowModal()
        aboutDialog.Destroy()

    def chooseShotFile(self, evt):
        """Choose a file to output the shotlog data to"""
        fileDialog = wxFileDialog(self, "Choose Shotlog Output File", style=wxSAVE)
        result = fileDialog.ShowModal()

        if result == wxID_OK:
            fileName = fileDialog.GetPath()

            # Now replace the control with the new pathname
            self.shotlogText.SetValue(fileName)

        fileDialog.Destroy()

    def shotTimeClicked(self, evt):
        """
        User clicked button - wishes to pick the shot time input file and
        describe the data - pop up this dialog
        """

        valueTypes = shotTimeValueTypes
        
        shotDialog = TextDataDialog(self, "Describe Shot Time Data", \
                                    fileio.shotTimeRegExpList, \
                                    valueTypes, self.shotTimeInfo)
        
        result = shotDialog.ShowModal()

        if result == wxID_OK:
            self.shotTimeInfo = shotDialog.getResult()

        shotDialog.Destroy()
        
        
    def navClicked(self, evt):
        """User wishes to describe the navigation data (file and format)"""

        valueTypes = navValueTypes
        
        navDialog = TextDataDialog(self, "Describe Navigation Data", fileio.navRegExpList, \
                                    valueTypes, self.navInfo)
        result = navDialog.ShowModal()

        if result == wxID_OK:
            self.navInfo = navDialog.getResult()            
        
        navDialog.Destroy()


    def runClicked(self, evt):
        """Run the backend process - capture the output and display to the user"""

        if self.cfgFileName == None:
            dlg = wxMessageDialog(self, "Error: please save options to a specific file", \
                                  "No Target Options File", wxOK | wxICON_WARNING) 
            dlg.ShowModal()
            dlg.Destroy()
            return

        # Ensure that the file is saved (implicitly)
        self.menuSave(None)

        # This starts the conversion in another thread (so we don't block the UI).
        runThread = RunShottabThread(self.cfgFileName, self)
        runThread.start()
        
        self.runOutputDialog = TextViewDialog(self, "Running shottab creation process", \
            "Now running shottab creation process\n")
        self.runOutputDialog.Show(True)
        
        # See self.runComplete for when this operation ends

    def addRunOutput(self, event):
        self.runOutputDialog.addText(event.data)
        
    def runComplete(self, event):
        # Event called when the RunShottabThread is complete
        self.runOutputDialog.addText("Creation process complete")
        # Now wait for them to click ok
        self.runOutputDialog.ShowModal()
        self.runOutputDialog.Destroy()
        
        
class OutputUpdater:
    """
    A class which looks like a file, passed to shottabMain to log messages.
    These are forwarded back to the GUI.
    
    """
    def __init__(self, updateMethod):
        self.updateMethod = updateMethod
        
    def write(self, string):
        self.updateMethod(string)
        
    def flush(self):
        pass

class RunShottabThread(threading.Thread):
    """A helper to run the shottab process in a different thread"""
    def __init__(self, optionsFile, ownerFrame):
        threading.Thread.__init__(self)
        self.optionsFile = optionsFile
        self.ownerFrame = ownerFrame
        
    def run(self):
        # Overriding the value of sys.out - will replace this later
        oldout = sys.stdout
        sys.stdout = OutputUpdater(updateMethod=self.addText)
        shotCreate.shottabMain(self.optionsFile)
        sys.stdout = oldout
        self.ownerFrame.AddPendingEvent(CustomRunEvent(eventId=RUN_COMPLETE_EVENT_ID, \
                data=None))
                
    def addText(self, text):
        """This is called when the shottab module wants to output some text"""
        self.ownerFrame.AddPendingEvent(CustomRunEvent(eventId=RUN_ADD__OUTPUT_EVENT_ID, \
            data=text))
        
class CustomRunEvent(wxPyEvent):
    """A little event class so we can pass data back to the UI"""
    def __init__(self, eventId, data):
        wxPyEvent.__init__(self)
        self.SetEventType(eventId)
        self.data = data

# This is the mainline of the GUI - where execution really starts.  We simply 
# create an application instance - set up our main window and let the GUI event loop run.
if __name__ == "__main__":

    # Command line arguments (if any)
    args = sys.argv[1:]
    cfgFile = None

    if len(args) > 0:
        # only option we support now is options filename
        cfgFile = args[0]        
        print("using config file: %s" % cfgFile)
        
    app = wxPySimpleApp()
    frame = MainWindow(None, -1, "Shottab Options Editor, Version " + VERSION, \
                       cfgFile)
    frame.Show(TRUE)
    app.MainLoop()
