# This file contains routines for obtaining heading information given
# by a series of lat/long coordinates.  The coordinates are converted
# to UTM, and a heading is derived for each (except the first).
#
# A (two sided) running mean is then taken which takes the mean of a
# given number of headings around the current one.  The resultant
# value is then the interpolated heading for this lat/long coordinate.
#
#
#      UTM <--> LAT/LON ROUTINES
#      FROM FORMULAE PROVIDED BY UNB SCHOOL OF SURVEYING
#      FROM SURVEYS AND MAPPING DIVISION OF THE
#      LAND REGISTRATION AND INFORMATION SERVICE
#
# This is a python adaptation of a heading interpolation program written by
# Keith Louden, and integrated into the pyShottab framework.
#
#
#
# Dalhousie University - Department of Oceanography.
#
# Author: Dave LeBlanc, dleblanc@cs.dal.ca, October 2002.

from math import *

sumNo = 9

# Ellipsoidal constant parameters from clarke 1866 ellipsiod
# Taken verbatim from fortran code.
A = 6378206.4
RECIPF = 294.97898
K0 = 0.9996

DEGRAD = pi / 180
RADDEG = 180 / pi
    
F = 1.0 / RECIPF
B = A - A * F
E1 = (A * A - B * B) / (A * A)
E2 = (A * A - B * B) / (B * B)

#     ARC LENGTH CONSTANTS
A0 = 1.0 - 0.25 * E1 - 0.046875 * E1 * E1 - 0.01953125 * E1**3.0 - 175.0 / 16384.0 * \
     E1**4.0
A2 = 0.375 * (E1 + 0.25 * E1 * E1 + 0.1171875 * E1**3.0 -455.0 / 4096.0 * E1**4.0)
A4 = 0.05859375 * (E1 * E1 + 0.75 * E1**3.0 - 77.0 / 128.0 * E1**4.0)
A6 = 35.0 / 3072.0 * (E1**3.0 - 1.28125 * E1**4.0)
A8 = -315.0 / 131072.0 * E1**4.0

def medarc(phi):
    b = A0 * phi - A2 * sin(2.0 * phi) + A4 * sin(4.0 * phi) - \
        A6 * sin(6.0 * phi) + A8 * sin(8.0 * phi)
    return A * b

def footpt(north):
    """Calculates foot point latitude, returned as a float"""

    latft = floor(north * 90 / 10000000) + 1
    phi = (latft - 1) * DEGRAD
    arc = medarc(phi) * K0

    while (north < arc):
        latft = latft - 1
        phi = (latft) * DEGRAD
        arc = medarc(phi) * K0

    latft = latft - 1
    phift = latft
    phi = (latft) * DEGRAD
    arc = medarc(phi) * K0
    latft = latft + 1
    phi = latft * DEGRAD
    arc2 = medarc(phi) * K0
    phift = phift + (north - arc)/(arc2 - arc)
    phi = phift * DEGRAD
    arc3 = medarc(phi) * K0
    return (phift + (north - arc3)/(arc2 - arc))

    
def latLongToUTM(lat, lon, centralMeridian):
    """
    This routine converts from the given LAT/LON coordinate
    and central meridian to the UTM coordinate system.
    
    This has been pythonified in that variables have been changed to
    lowercase, whitespace has been introduced between operators,
    and it is a function returning the tuple (north, east),
    but is otherwise very close to the Fortran/C versions.
    """
    north = 0.0
    east = 0.0

    # since this is different in here and UTMToLatLog, just keep it local.
    FE = 0.0

    phi = lat / 57.29577951
    v = E1 * sin(phi)**2.0
    n = A / sqrt(1.0 - v)
    t = tan(phi)**2.0
    delta = (lon - centralMeridian) / 57.29577951
    n2 = E2 * cos(phi)**2.0
    s0 = medarc(phi)

    t1 = delta * delta / 2.0 * sin(phi) * cos(phi)
    t2 = delta**4.0 / 24.0 * sin(phi) * cos(phi)**3.0 * \
         (5.0 - t + 9.0 * n2 + 4.0 * n2 * n2)
    
    t3 = 61.0 - 58.0 * t + t * t + 270.0 * n2 - 330.0 * t * \
         n2 + 445.0 * n2 * n2 + 324.0 * n2**3.0 - 680.0 * \
         n2 * n2 * t + 88.0 * n2**4.0 - 600. * n2**3.0 * t - \
         192.0 * n2**4.0 * t
    
    t4 = 1385.0 - 3111.0 * t + 543.0 * t * t - t**3.0
    north = K0 * s0 + K0 * n * \
            (t1 + t2 + delta**6.0 / 720.0 * sin(phi) * cos(phi)**5.0 *
            t3 + delta**8.0 / 40320.0 * sin(phi) * cos(phi)**7.0 * t4)
    
    t1 = delta * cos(phi) + (delta * cos(phi))**3.0 / 6.0 * (1 - t + n2)

    t2 = 5.0 -18.0 * t + t * t + 14.0 * n2 - 58.0 * t * n2 + 13.0 * n2 * \
         n2 + 4.0 * n2**3.0 - 64.0 * n2 * n2 * t - 24.0 * n2**3.0 * t
    
    t3 = (61.0 - 479.0 * t + 179 * t * t - t**3.0) / 5040.0

    east = K0 * n * \
           (t1 + (delta * cos(phi))**5.0 / 120.0 * t2 + \
            (delta * cos(phi))**7.0 * t3)
    
    east += 500000.0

    return (north, east)

def UTMToLatLong(north, east, centralMeridian):
    """
    Return (lat, lon) given the UTM input and central meridian.
    """

    FE = 500000

    phift = footpt(north)
    phift = phift * DEGRAD
    y = north/K0
    x = (east-FE)/K0
    t = tan(phift)

    # Meridian radius of curvature at foot point latitude
    m = A * (1 - E1) / (1 - E1 * sin(phift) * sin(phift)) ** 1.5
    
    # Prime vertical radius of curvature at foot point latitude
    n1 = A / sqrt(1 - E1 * sin(phift) * sin(phift))
    n = E2 * cos(phift) * cos(phift)

    # The rest of this code is very dense, so I'll leave out the whitespace
    t1 =-t*x/m*x/n1/2+t*x/m*x/n1*x/n1*x/n1/24*(5+3*t*t+n+4*n*n+9*n*t*t)
    
    t2 =t*x/m*x/n1*x/n1*x/n1*x/n1*x/n1/720* \
         (61-90*t*t+46*n+45*pow(t,4)-252*t*t*n+3*n*n+100*pow(n,3)-66*t*t*n*n-90* \
          pow(t,4)*n+88*pow(n,4)+225*pow(t,4)*n*n+84*t*t*pow(n,3)-192*t*t*pow(n,4))
    
    t3 =t*x/m*x/n1*x/n1*x/n1*x/n1*x/n1*x/n1*x/n1/40320* \
        (1385+3633*t*t+4095*pow(t,4)+1575*pow(t,6))
    
    lat =(phift+t1-t2+t3)*RADDEG
    
    t1 =x/n1-x/n1*x/n1*x/n1/6*(1+2*t*t+n)
    t2 =5+6*n+28*t*t-3*n*n+8*t*t*n+24*pow(t,4)-4*pow(n,3)+4*t*t*n*n+24*t*t*pow(n,3)
    t3 =61+662*t*t+1320*pow(t,4)+720*pow(t,6)

    lon =centralMeridian+(t1+x/n1*x/n1*x/n1*x/n1*x/n1/120*t2-x/n1*x/n1*x/n1*x/n1*x/n1*x/n1
            *x/n1/5040*t3)/cos(phift)*RADDEG

    return (lat, lon)
    
def getHeading(fromLat, fromLon, toLat, toLon, centralMeridian):
    """
    Get a heading from fromCoord (lat/long) to toCoord (lat/long)
    given the supplied central meridian.
    """
    (fromNorth, fromEast) = latLongToUTM(fromLat, fromLon, centralMeridian)
    (toNorth, toEast) = latLongToUTM(toLat, toLon, centralMeridian)

    # Convert this to a heading and return it
    return atan2(toEast - fromEast, toNorth - fromNorth) * RADDEG
