#!/usr/bin/env python
#
# Unit Tests - anything in this module that implements unittest.TestCase is run
#

import sys, os
import unittest

from types import *
import StringIO

from shottab.options import options
import shottab.core.shotCreate as shotCreate
import shottab.fileio.fileio as fileio

class OptionsTestLoader(options.Options):
    def __init__(self, optionsFile):
        # Initialize with the opposite of what we want
        self.falseBoolValue = True
        self.falseBoolValueInt = True

        self.trueBoolValue = False
        self.trueBoolValueInt = False

        self.intValue = 321
        self.longValue = 1010
        self.floatValue = 1.23456

        options.Options.__init__(self, optionsFile)

class OptionsTests(unittest.TestCase):
    def testCustomBool(self):
        self.assertEquals(False, options.customBool("0"))
        self.assertEquals(False, options.customBool("False"))
        
        self.assertEquals(True, options.customBool("1"))
        self.assertEquals(True, options.customBool("True"))
        
        try:
            self.assertEquals(True, options.customBool("blargh!"))
            self.fail("expected a InvalidValueError exception")
        except options.InvalidValueError:
            # expected
            pass
            
    def testOptionsFile(self):
        # Load with no options file - ensure default stays
        options = OptionsTestLoader(None)
        self.assertEquals(BooleanType, type(options.trueBoolValue))
        
        options = OptionsTestLoader("shottab/tests/optionsTest.ini")
        
        self.assertEquals(BooleanType, type(options.trueBoolValue))
        self.assertEquals(True, options.trueBoolValue)
        self.assertEquals(True, options.trueBoolValueInt)
        
        self.assertEquals(False, options.falseBoolValue)
        self.assertEquals(False, options.falseBoolValueInt)

        self.assertEquals(123, options.intValue)
        
        self.assertEquals(18446744073709551616L, options.longValue)
        
        self.assertEquals(1.23456, options.floatValue)

class TimeTester(unittest.TestCase):
    def testGetNanoTime(self):
        expectedTime = 273906789000000L
        time = shotCreate.getNanoTime(3, 4, 5, 6.789)
        self.assertEquals(expectedTime, time)

    def testGetNanoTime(self):
        expectedTime = 31626059999900000L
        time = shotCreate.getNanoTime(365, 24, 60, 59.9999)
        self.assertEquals(expectedTime, time)

    def testNanoToDayHourMinSec(self):
        expectedTime = 3, 4, 5, 6.789
        timeTuple = shotCreate.nanoToDayHourMinSec(273906789000000L)
        for idx in range(len(timeTuple)):
            self.assertEquals(expectedTime[idx], timeTuple[idx])

class FileIoTest(unittest.TestCase):
    baseDir = "shottab/tests/"
    def testSimpleLineReader(self):
        filename = self.baseDir + "simpleData.txt"
        valueTypes = (("first", str), ("second", str), ("third", str))
        regexp = """
        	(?P<first>\w*)
        	\s*
        	(?P<second>\w*)
        	\s*
        	(?P<third>\w*)"""
            
        reader = fileio.LazyLineReader(regexp, valueTypes, filename, 0)

        # Read the 1st line
        line1 = reader[0]
        self.assertNotEquals(None, line1)
        self.assertEquals("line1col1", line1[0])
        self.assertEquals("line1col2", line1[1])
        del line1

        # now the second
        line2 = reader[1]
        self.assertNotEquals(None, line2)
        self.assertEquals("line2col1", line2[0])
        self.assertEquals("line2col2", line2[1])
        del line2
        
        # ensure the 1st is still sane
        line1 = reader[0]
        self.assertNotEquals(None, line1)
        self.assertEquals("line1col1", line1[0])
        self.assertEquals("line1col2", line1[1])
        del line1
        
        # Ensure we get an eof error on the 3rd line
        try:
            line3 = reader[2]
            self.fail("expected an end of file error")
        except fileio.LazyEOF:
            pass

    def testTimeReader(self):
        """Test TimeReader with a custom regular expression - no partialSec"""
        filename = self.baseDir + "simpleTimeData.txt"

        regExp = """
           	(?P<day>\d*)
        	\s*
        	(?P<hour>\d*)
        	\s*
        	(?P<min>\d*) 
        	\s*
        	(?P<sec>[\d.]*) 
        	"""
            
        reader = shotCreate.TimeReader(regExp, shotCreate.navValueTypes, filename, 0)
            
        # Read the 1st line
        line1 = reader[0]
        self.assertNotEquals(None, line1)
        self.assertEquals(1, line1[shotCreate.NAV_DAY_IDX])
        self.assertEquals(2, line1[shotCreate.NAV_HOUR_IDX])
        self.assertEquals(3, line1[shotCreate.NAV_MIN_IDX])
        self.assertEquals(4.56, line1[shotCreate.NAV_SEC_IDX])

        # Ensure the time value is correct
        self.assertEquals(93784560000000L, line1[shotCreate.NAV_TIME_IDX])                          

    def testPartialSecTimeReader(self):
        """testPartialSecTimeReader: test TimeReader with partialSec defined"""
        filename = self.baseDir + "partialSecTimeData.txt"
        reader = shotCreate.TimeReader(fileio.shotTimeRegExpList[0][1], 
            shotCreate.shotTimeValueTypes, filename, 0)

        # Read the 1st line
        line1 = reader[0]
        self.assertNotEquals(None, line1)
        self.assertEquals(1, line1[shotCreate.SHOT_DAY_IDX])
        self.assertEquals(2, line1[shotCreate.SHOT_HOUR_IDX])
        self.assertEquals(3, line1[shotCreate.SHOT_MIN_IDX])
        self.assertEquals(4.0, line1[shotCreate.SHOT_SEC_IDX])
        self.assertEquals("05678", line1[shotCreate.SHOT_PARTIAL_SEC_IDX])
        
        # Ensure the time field is set correctly.
        self.assertEquals(93784056780000, line1[shotCreate.SHOT_TIME_IDX])
        